import uuid

from datetime import datetime
from typing import List, Optional

import brewman.schemas.input as schema_in
import brewman.schemas.voucher as output

from fastapi import APIRouter, Depends, File, HTTPException, Request, Security, status
from sqlalchemy import distinct, select
from sqlalchemy.exc import SQLAlchemyError
from sqlalchemy.orm import Session

from ..core.security import get_current_active_user as get_user
from ..core.session import get_date, set_date
from ..db.session import SessionFuture
from ..models.account_base import AccountBase
from ..models.journal import Journal
from ..models.validations import check_journals_are_valid
from ..models.voucher import Voucher
from ..models.voucher_type import VoucherType
from ..schemas.account import AccountLink
from ..schemas.blank_voucher_info import BlankVoucherInfo
from ..schemas.user import UserToken
from . import get_lock_info
from .db_image import save_files, update_files
from .voucher import blank_voucher, check_voucher_edit_allowed, voucher_info


router = APIRouter()


@router.post("", response_model=output.Voucher)
def save_route(
    request: Request,
    data: schema_in.JournalIn = Depends(schema_in.JournalIn.load_form),
    i: List[bytes] = File(None),
    t: List[bytes] = File(None),
    user: UserToken = Security(get_user, scopes=["journal"]),
) -> output.Voucher:
    try:
        with SessionFuture() as db:
            item: Voucher = save(data, user, db)
            db.flush()
            check_journals_are_valid(item)
            save_files(item.id, i, t, db)
            db.commit()
            set_date(data.date_.strftime("%d-%b-%Y"), request.session)
            info = voucher_info(item, db)
            return info
    except SQLAlchemyError as e:
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail=str(e),
        )


def save(data: schema_in.JournalIn, user: UserToken, db: Session) -> Voucher:
    account_types = (
        db.execute(
            select(distinct(AccountBase.type_id)).where(AccountBase.id.in_([dj.account.id_ for dj in data.journals]))
        )
        .scalars()
        .all()
    )
    allowed, message = get_lock_info([data.date_], VoucherType[data.type_.replace(" ", "_").upper()], account_types, db)
    if not allowed:
        raise HTTPException(
            status_code=status.HTTP_423_LOCKED,
            detail=message,
        )
    voucher = Voucher(
        date=data.date_,
        narration=data.narration,
        is_starred=data.is_starred,
        user_id=user.id_,
        voucher_type=VoucherType[data.type_.replace(" ", "_").upper()],
    )
    db.add(voucher)
    for item in data.journals:
        account: AccountBase = db.execute(select(AccountBase).where(AccountBase.id == item.account.id_)).scalar_one()
        journal = Journal(
            amount=round(item.amount, 2),
            debit=item.debit,
            account_id=account.id,
            cost_centre_id=account.cost_centre_id,
        )
        voucher.journals.append(journal)
        db.add(journal)
    return voucher


@router.put("/{id_}", response_model=output.Voucher)
def update_route(
    id_: uuid.UUID,
    request: Request,
    data: schema_in.JournalIn = Depends(schema_in.JournalIn.load_form),
    i: List[bytes] = File(None),
    t: List[bytes] = File(None),
    user: UserToken = Security(get_user, scopes=["journal"]),
) -> output.Voucher:
    try:
        with SessionFuture() as db:
            item: Voucher = update_voucher(id_, data, user, db)
            check_journals_are_valid(item)
            update_files(item.id, data.files, i, t, db)
            db.commit()
            set_date(data.date_.strftime("%d-%b-%Y"), request.session)
            return voucher_info(item, db)
    except SQLAlchemyError as e:
        db.rollback()
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail=str(e),
        )
    except Exception:
        db.rollback()
        raise


def update_voucher(id_: uuid.UUID, data: schema_in.JournalIn, user: UserToken, db: Session) -> Voucher:
    voucher: Voucher = db.execute(select(Voucher).where(Voucher.id == id_)).scalar_one()
    account_types = (
        db.execute(
            select(distinct(AccountBase.type_id)).where(
                AccountBase.id.in_(
                    [dj.account.id_ for dj in data.journals] + [vj.account_id for vj in voucher.journals]
                )
            )
        )
        .scalars()
        .all()
    )
    allowed, message = get_lock_info([voucher.date, data.date_], voucher.voucher_type, account_types, db)
    if not allowed:
        raise HTTPException(
            status_code=status.HTTP_423_LOCKED,
            detail=message,
        )

    check_voucher_edit_allowed(voucher, user)
    voucher.date = data.date_
    voucher.is_starred = data.is_starred
    voucher.narration = data.narration
    voucher.user_id = user.id_
    voucher.posted = False
    voucher.last_edit_date = datetime.utcnow()

    for i in range(len(voucher.journals), 0, -1):
        item = voucher.journals[i - 1]
        index = next((idx for (idx, d) in enumerate(data.journals) if d.id_ == item.id), None)
        if index is not None:
            new_item = data.journals.pop(index)
            account = db.execute(select(AccountBase).where(AccountBase.id == new_item.account.id_)).scalar_one()
            item.debit = new_item.debit
            item.amount = round(new_item.amount, 2)
            item.account_id = account.id
            item.cost_centre_id = account.cost_centre_id
        else:
            voucher.journals.remove(item)
    for new_item in data.journals:
        account = db.execute(select(AccountBase).where(AccountBase.id == new_item.account.id_)).scalar_one()
        journal = Journal(
            amount=new_item.amount,
            debit=round(new_item.debit, 2),
            account_id=account.id,
            cost_centre_id=account.cost_centre_id,
        )
        db.add(journal)
        voucher.journals.append(journal)
    return voucher


@router.get("/{id_}", response_model=output.Voucher)
def get_id(
    id_: uuid.UUID,
    user: UserToken = Security(get_user, scopes=["journal"]),
) -> output.Voucher:
    try:
        with SessionFuture() as db:
            item: Voucher = db.execute(select(Voucher).where(Voucher.id == id_)).scalar_one()
            return voucher_info(item, db)
    except SQLAlchemyError as e:
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail=str(e),
        )


@router.get("", response_model=output.Voucher)
def show_blank(
    request: Request,
    a: Optional[uuid.UUID] = None,
    user: UserToken = Security(get_user, scopes=["journal"]),
) -> output.Voucher:
    if request.scope.get("path") == "/api/payment":
        type_ = VoucherType.PAYMENT
    elif request.scope.get("path") == "/api/receipt":
        type_ = VoucherType.RECEIPT
    else:
        type_ = VoucherType.JOURNAL

    additional_info = BlankVoucherInfo(date=get_date(request.session), type=type_)
    if a:
        additional_info.account = AccountLink(id=a)
    with SessionFuture() as db:
        return blank_voucher(additional_info, db)
