from datetime import date, datetime, time, timedelta
from typing import List, Optional, Tuple

import brewman.schemas.entries as schemas

from fastapi import APIRouter, Depends, Security
from sqlalchemy import desc, not_, or_, select
from sqlalchemy.orm import Session, contains_eager, joinedload
from sqlalchemy.sql.functions import count

from ...core.security import get_current_active_user as get_user
from ...db.session import SessionFuture
from ...models.journal import Journal
from ...models.voucher import Voucher
from ...models.voucher_type import VoucherType
from ...schemas.user import UserToken
from ...schemas.user_link import UserLink
from . import report_finish_date, report_start_date


router = APIRouter()


def build_report(
    start_date: Optional[date],
    finish_date: Optional[date],
    posted: Optional[bool],
    issue: bool,
    page_size: int,
    page_index: int,
    active_sort: str,
    sort_direction: str,
    db: Session,
) -> Tuple[int, List[Voucher]]:
    query = (
        select(Voucher)
        .join(Voucher.user)
        .options(
            joinedload(Voucher.user, innerjoin=True),
            joinedload(Voucher.journals, innerjoin=True).joinedload(Journal.account, innerjoin=True),
            contains_eager(Voucher.user),
            contains_eager(Voucher.journals, Journal.account),
        )
    )
    sq = select(Voucher.id)
    counts = select(count(Voucher.id))
    if start_date is not None:
        sd = datetime.combine(start_date, time(5, 30))
        sq = sq.where(or_(Voucher.creation_date >= sd, Voucher.last_edit_date >= sd))
        counts = counts.where(or_(Voucher.creation_date >= sd, Voucher.last_edit_date >= sd))
    if finish_date is not None:
        fd = datetime.combine(finish_date, time(5, 30)) + timedelta(days=1)
        sq = sq.where(or_(Voucher.creation_date <= fd, Voucher.last_edit_date <= fd))
        counts = counts.where(or_(Voucher.creation_date <= fd, Voucher.last_edit_date <= fd))
    if posted is not None:
        sq = sq.where(Voucher.posted == posted)
        counts = counts.where(Voucher.posted == posted)
    if issue is False:
        sq = sq.where(not_(Voucher.voucher_type.in_([VoucherType.ISSUE, VoucherType.CLOSING_STOCK])))
        counts = counts.where(
            not_(Voucher.voucher_type.in_([VoucherType.ISSUE, VoucherType.CLOSING_STOCK])),
        )
    if active_sort == "date":
        if sort_direction == "desc":
            sq = sq.order_by(desc(Voucher.date), desc(Voucher.last_edit_date))
            query = query.order_by(desc(Voucher.date), desc(Voucher.last_edit_date))
        else:
            sq = sq.order_by(Voucher.date, Voucher.last_edit_date)
            query = query.order_by(Voucher.date, Voucher.last_edit_date)
    if active_sort == "user":
        if sort_direction == "desc":
            sq = sq.order_by(desc(Voucher.last_edit_date))
            query = query.order_by(desc(Voucher.last_edit_date))
        else:
            sq = sq.order_by(Voucher.last_edit_date)
            query = query.order_by(Voucher.last_edit_date)
    if page_size:
        sq = sq.limit(page_size)
    if page_index:
        sq = sq.offset(page_size * page_index)
    return db.execute(counts).scalar_one(), db.execute(query.where(Voucher.id.in_(sq))).unique().scalars().all()


@router.get("", response_model=schemas.Report)
def report_data(
    start_date: Optional[date] = Depends(report_start_date),
    finish_date: Optional[date] = Depends(report_finish_date),
    p: Optional[bool] = None,  # Posted
    ps: Optional[int] = 50,  # Page Size
    pi: Optional[int] = 0,  # Page Index
    a: Optional[str] = "date",  # Active Sort
    d: Optional[str] = "desc",  # Sort Direction
    i: Optional[bool] = False,  # Show issue vouchers
    user: UserToken = Security(get_user, scopes=["ledger"]),
) -> schemas.Report:
    assert i is not None and ps is not None and pi is not None and a is not None and d is not None
    with SessionFuture() as db:
        counts, report = build_report(start_date, finish_date, p, i, ps, pi, a, d, db)
        return schemas.Report(
            counts=counts,
            report=[
                schemas.Entries(
                    id=voucher.id,
                    date=voucher.date,
                    url=[
                        "/",
                        voucher.voucher_type.name.replace("_", "-").lower(),
                        str(voucher.id),
                    ],
                    type=voucher.voucher_type.name.replace("_", " ").title(),
                    posted=voucher.posted,
                    narration=voucher.narration,
                    debitNames=[x.account.name for x in voucher.journals if x.debit == 1],
                    creditNames=[x.account.name for x in voucher.journals if x.debit != 1],
                    amount=sum(x.amount for x in voucher.journals if x.debit != 1),
                    creationDate=voucher.creation_date,
                    lastEditDate=voucher.last_edit_date,
                    user=UserLink(id=voucher.user.id, name=voucher.user.name),
                )
                for voucher in report
            ],
        )
