import uuid

from datetime import date, datetime
from decimal import Decimal
from typing import Any, Optional, Union

from pydantic import BaseModel, Field, validator

from . import to_camel
from .account import AccountBase
from .cost_centre import CostCentreLink


class EmployeeIn(AccountBase):
    designation: str
    salary: int = Field(ge=0)
    points: Decimal = Field(ge=0, lt=1000, multiple_of=0.01)
    joining_date: date
    leaving_date: Optional[date]

    @validator("joining_date", pre=True)
    def parse_joining_date(cls, value: Union[date, str]) -> date:
        if isinstance(value, date):
            return value
        return datetime.strptime(value, "%d-%b-%Y").date()

    @validator("leaving_date", pre=True)
    def parse_leaving_date(cls, value: Union[None, date, str]) -> Optional[date]:
        if isinstance(value, date):
            return value
        if value is None:
            return value
        return datetime.strptime(value, "%d-%b-%Y").date()

    @validator("leaving_date")
    def leaving_date_more_than_joining_date(cls, v: Optional[date], values: dict, **kwargs: Any) -> Optional[date]:
        if values["is_active"]:
            return None
        if v < values["joining_date"]:
            raise ValueError("Leaving Date cannot be less than Joining Date")
        return v

    class Config:
        anystr_strip_whitespace = True
        alias_generator = to_camel
        json_encoders = {datetime: lambda v: v.strftime("%d-%b-%Y %H:%M"), date: lambda v: v.strftime("%d-%b-%Y")}


class Employee(EmployeeIn):
    id_: uuid.UUID
    code: int
    is_fixture: bool


class EmployeeBlank(EmployeeIn):
    name: str

    class Config:
        anystr_strip_whitespace = True
        alias_generator = to_camel
        json_encoders = {datetime: lambda v: v.strftime("%d-%b-%Y %H:%M"), date: lambda v: v.strftime("%d-%b-%Y")}


class EmployeeLink(BaseModel):
    id_: uuid.UUID = Field(...)
    name: Optional[str]
    designation: Optional[str]
    cost_centre: CostCentreLink

    class Config:
        alias_generator = to_camel
