import { AfterViewInit, Component, ElementRef, OnInit, OnDestroy, ViewChild } from '@angular/core';
import { FormControl, FormGroup } from '@angular/forms';
import { MatAutocompleteSelectedEvent } from '@angular/material/autocomplete';
import { MatDialog } from '@angular/material/dialog';
import { ActivatedRoute, Router } from '@angular/router';
import { Hotkey, HotkeysService } from 'angular2-hotkeys';
import { round } from 'mathjs';
import * as moment from 'moment';
import { BehaviorSubject, Observable, of as observableOf } from 'rxjs';
import { debounceTime, distinctUntilChanged, map, switchMap } from 'rxjs/operators';

import { AuthService } from '../auth/auth.service';
import { Account } from '../core/account';
import { AccountBalance } from '../core/account-balance';
import { AccountService } from '../core/account.service';
import { Batch } from '../core/batch';
import { BatchService } from '../core/batch.service';
import { DbFile } from '../core/db-file';
import { Inventory } from '../core/inventory';
import { ToasterService } from '../core/toaster.service';
import { User } from '../core/user';
import { Voucher } from '../core/voucher';
import { VoucherService } from '../core/voucher.service';
import { ConfirmDialogComponent } from '../shared/confirm-dialog/confirm-dialog.component';
import { ImageDialogComponent } from '../shared/image-dialog/image-dialog.component';
import { ImageService } from '../shared/image.service';
import { MathService } from '../shared/math.service';

import { PurchaseReturnDataSource } from './purchase-return-datasource';
import { PurchaseReturnDialogComponent } from './purchase-return-dialog.component';

@Component({
  selector: 'app-purchase-return',
  templateUrl: './purchase-return.component.html',
  styleUrls: ['./purchase-return.component.css'],
})
export class PurchaseReturnComponent implements OnInit, AfterViewInit, OnDestroy {
  @ViewChild('accountElement', { static: true }) accountElement?: ElementRef;
  @ViewChild('batchElement', { static: true }) batchElement?: ElementRef;
  @ViewChild('dateElement', { static: true }) dateElement?: ElementRef;
  public inventoryObservable = new BehaviorSubject<Inventory[]>([]);
  dataSource: PurchaseReturnDataSource = new PurchaseReturnDataSource(this.inventoryObservable);
  form: FormGroup<{
    date: FormControl<Date>;
    account: FormControl<Account | string | null>;
    amount: FormControl<number>;
    addRow: FormGroup<{
      batch: FormControl<Batch | string | null>;
      quantity: FormControl<string>;
    }>;
    narration: FormControl<string>;
  }>;

  voucher: Voucher = new Voucher();
  batch: Batch | null = null;
  accBal: AccountBalance | null = null;

  displayedColumns = ['product', 'quantity', 'rate', 'tax', 'discount', 'amount', 'action'];

  accounts: Observable<Account[]>;
  batches: Observable<Batch[]>;

  constructor(
    private route: ActivatedRoute,
    private router: Router,
    private dialog: MatDialog,
    private hotkeys: HotkeysService,
    private toaster: ToasterService,
    public auth: AuthService,
    private math: MathService,
    public image: ImageService,
    private ser: VoucherService,
    private batchSer: BatchService,
    private accountSer: AccountService,
  ) {
    this.form = new FormGroup({
      date: new FormControl(new Date(), { nonNullable: true }),
      account: new FormControl<Account | string | null>(null),
      amount: new FormControl({ value: 0, disabled: true }, { nonNullable: true }),
      addRow: new FormGroup({
        batch: new FormControl<Batch | string | null>(null),
        quantity: new FormControl<string>('', { nonNullable: true }),
      }),
      narration: new FormControl('', { nonNullable: true }),
    });
    // Listen to Account Autocomplete Change
    this.accounts = this.form.controls.account.valueChanges.pipe(
      map((x) => (x instanceof Account ? x.name : x)),
      map((x) => (x !== null && x.length >= 1 ? x : null)),
      debounceTime(150),
      distinctUntilChanged(),
      switchMap((x) => (x === null ? observableOf([]) : this.accountSer.autocomplete(x))),
    );
    // Listen to Batch Autocomplete Change
    this.batches = this.form.controls.addRow.controls.batch.valueChanges.pipe(
      map((x) => (x instanceof Batch ? x.name : x)),
      map((x) => (x !== null && x.length >= 1 ? x : '')),
      debounceTime(150),
      distinctUntilChanged(),
      switchMap((x) =>
        this.batchSer.autocomplete(moment(this.form.value.date).format('DD-MMM-YYYY'), x),
      ),
    );
  }

  ngOnInit() {
    this.route.data.subscribe((value) => {
      const data = value as { voucher: Voucher };

      this.loadVoucher(data.voucher);
    });
    this.hotkeys.add(
      new Hotkey(
        'f2',
        (): boolean => {
          setTimeout(() => {
            if (this.dateElement) {
              this.dateElement.nativeElement.focus();
            }
          }, 0);
          return false; // Prevent bubbling
        },
        ['INPUT', 'SELECT', 'TEXTAREA'],
      ),
    );
    this.hotkeys.add(
      new Hotkey(
        'ctrl+s',
        (): boolean => {
          if (this.canSave()) {
            this.save();
          }
          return false; // Prevent bubbling
        },
        ['INPUT', 'SELECT', 'TEXTAREA'],
      ),
    );
    this.hotkeys.add(
      new Hotkey(
        'ctrl+p',
        (): boolean => {
          if (this.voucher.id && !this.voucher.posted && this.auth.allowed('post-vouchers')) {
            this.post();
          }
          return false; // Prevent bubbling
        },
        ['INPUT', 'SELECT', 'TEXTAREA'],
      ),
    );
  }

  ngAfterViewInit() {
    this.focusAccount();
  }

  ngOnDestroy() {
    this.hotkeys.reset();
  }

  loadVoucher(voucher: Voucher) {
    this.voucher = voucher;
    this.form.setValue({
      date: moment(this.voucher.date, 'DD-MMM-YYYY').toDate(),
      account: this.voucher.vendor || null,
      amount: Math.abs(this.voucher.inventories.map((x) => x.amount).reduce((p, c) => p + c, 0)),
      addRow: {
        batch: '',
        quantity: '',
      },
      narration: this.voucher.narration,
    });
    this.dataSource = new PurchaseReturnDataSource(this.inventoryObservable);
    this.updateView();
  }

  focusAccount() {
    setTimeout(() => {
      if (this.accountElement) {
        this.accountElement.nativeElement.focus();
      }
    }, 0);
  }

  addRow() {
    const formValue = this.form.value.addRow!;
    const quantity = this.math.parseAmount(formValue.quantity, 2);
    if (this.batch === null || quantity <= 0 || this.batch.quantityRemaining < quantity) {
      return;
    }
    const old = this.voucher.inventories.find((x) => x.batch.id === (this.batch as Batch).id);
    if (old !== undefined) {
      this.toaster.show('Danger', 'Product already added');
      return;
    }
    this.voucher.inventories.push(
      new Inventory({
        quantity,
        rate: this.batch.rate,
        tax: this.batch.tax,
        discount: this.batch.discount,
        amount: quantity * this.batch.rate * (1 + this.batch.tax) * (1 - this.batch.discount),
        batch: this.batch,
      }),
    );
    this.resetAddRow();
    this.updateView();
  }

  resetAddRow() {
    this.form.controls.addRow.reset();
    this.batch = null;
    setTimeout(() => {
      if (this.batchElement) {
        this.batchElement.nativeElement.focus();
      }
    }, 0);
  }

  updateView() {
    this.inventoryObservable.next(this.voucher.inventories);
    this.form.controls.amount.setValue(
      round(Math.abs(this.voucher.inventories.map((x) => x.amount).reduce((p, c) => p + c, 0)), 2),
    );
  }

  editRow(row: Inventory) {
    const dialogRef = this.dialog.open(PurchaseReturnDialogComponent, {
      width: '750px',
      data: {
        inventory: { ...row },
        date: moment(this.form.value.date).format('DD-MMM-YYYY'),
      },
    });

    dialogRef.afterClosed().subscribe((result: boolean | Inventory) => {
      if (!result) {
        return;
      }
      const j = result as Inventory;
      if (
        j.batch.sku.id !== row.batch.sku.id &&
        this.voucher.inventories.filter((x) => x.batch.sku.id === j.batch.sku.id).length
      ) {
        return;
      }
      Object.assign(row, j);
      this.updateView();
    });
  }

  deleteRow(row: Inventory) {
    this.voucher.inventories.splice(this.voucher.inventories.indexOf(row), 1);
    this.updateView();
  }

  canSave() {
    if (!this.voucher.id) {
      return true;
    }
    if (this.voucher.posted && this.auth.allowed('edit-posted-vouchers')) {
      return true;
    }
    return (
      this.voucher.user.id === (this.auth.user as User).id ||
      this.auth.allowed("edit-other-user's-vouchers")
    );
  }

  post() {
    this.ser.post(this.voucher.id as string).subscribe(
      (result) => {
        this.loadVoucher(result);
        this.toaster.show('Success', 'Voucher Posted');
      },
      (error) => {
        this.toaster.show('Danger', error);
      },
    );
  }

  save() {
    const voucher: Voucher = this.getVoucher();
    this.ser.saveOrUpdate(voucher).subscribe(
      (result) => {
        this.toaster.show('Success', '');
        if (voucher.id === result.id) {
          this.loadVoucher(result);
        } else {
          this.router.navigate(['/purchase-return', result.id]);
        }
      },
      (error) => {
        this.toaster.show('Danger', error);
      },
    );
  }

  getVoucher(): Voucher {
    const formModel = this.form.value;
    this.voucher.date = moment(formModel.date).format('DD-MMM-YYYY');
    if (formModel.account instanceof Account) {
      this.voucher.vendor = formModel.account;
    }
    this.voucher.narration = formModel.narration!;
    return this.voucher;
  }

  delete() {
    this.ser.delete(this.voucher.id as string).subscribe(
      () => {
        this.toaster.show('Success', '');
        this.router.navigate(['/purchase-return'], { replaceUrl: true });
      },
      (error) => {
        this.toaster.show('Danger', error);
      },
    );
  }

  confirmDelete(): void {
    const dialogRef = this.dialog.open(ConfirmDialogComponent, {
      width: '250px',
      data: { title: 'Delete Voucher?', content: 'Are you sure? This cannot be undone.' },
    });

    dialogRef.afterClosed().subscribe((result: boolean) => {
      if (result) {
        this.delete();
      }
    });
  }

  displayFn(item?: Account | Batch): string {
    return item ? item.name : '';
  }

  accountSelected(event: MatAutocompleteSelectedEvent): void {
    this.form.controls.account.setValue(event.option.value);
  }

  batchSelected(event: MatAutocompleteSelectedEvent): void {
    this.batch = event.option.value;
  }

  zoomImage(file: DbFile) {
    this.dialog.open(ImageDialogComponent, {
      width: '750px',
      data: file.resized,
    });
  }

  deleteImage(file: DbFile) {
    const index = this.voucher.files.indexOf(file);
    this.voucher.files.splice(index, 1);
  }
}
